(* 
app load ["Int", "Csiga", "Listsort", "Regex", "Timer"]

val f = "test01d.txt"

*)

structure KCsiga :> KCsiga =
    (* dp02s nhf v1.0 -- 2002. prilis 17. Hank Dvid *)
struct

(* A bemenetet, ill. a kimenetet a terminlra irnythatjuk Unix alatt
   a /dev/stdin, ill. a /dev/stdout, MS-DOS alatt mindkt esetben a con
   llomnynv megadsval.
*)

exception KCsiga of string;

fun helyes_feladvany_leiro(Meret, Ciklus, Adottak) =
    let
	fun helyes_adott_elem(Sor, Oszlop, Ertek) = 0 < Sor andalso
	      Sor <= Meret andalso 0 < Oszlop andalso Oszlop <= Meret andalso
	      0 < Ertek andalso Ertek <= Ciklus;

	fun lexikografikusan_hasonlit ((x1,y1,_), (x2,y2,_)) =
	      if      x1 < x2 orelse x1 = x2 andalso y1 < y2 then LESS
	      else if x1 > x2 orelse x1 = x2 andalso y1 > y2 then GREATER
	      else EQUAL

	fun szomszedok_kulonbozok ((x1,y1,_)::(x2y2s as (x2,y2,_)::xys)) =
	   (x1 <> x2 orelse y1 <> y2) andalso szomszedok_kulonbozok x2y2s
	  | szomszedok_kulonbozok _ = true
    in
	0 < Meret andalso 0 < Ciklus andalso Ciklus <= Meret andalso
          (* minden eleme helyes az Adottak listnak: *)
        (List.all (fn x => x) (map helyes_adott_elem Adottak)) andalso
          (* az Adottak lista lexikografikusan rendezett: *)
	Listsort.sorted lexikografikusan_hasonlit Adottak andalso
	  (* Az Adottak lista minden eleme klnbz (Sor,Oszlop) 
	     koordintra vonatkozik (a rendezettsg miatt elegend
             ezt a szomszdos elemekre kiktni): *)
	szomszedok_kulonbozok Adottak
    end

(* csiga_be f = az f szveges llomnybl beolvasott adatokbl kpzett
   feladvnyler.
   csiga_be : string -> TCsiga.feladvany_leiro
 *)
fun csiga_be f =
    let
	fun harmas [a,b,c] = (a,b,c)
	  | harmas _ =
	    raise KCsiga("Szintaktikusan hibs adat a(z) " ^ f ^ " llomnyban.")

	val fuzerbol_adottak = harmas                           o
	                       (List.mapPartial Int.fromString) o
			       (String.tokens Char.isSpace)

	fun adottak_be s =
	    if TextIO.endOfStream s then []
	    else fuzerbol_adottak (TextIO.inputLine s) :: adottak_be s

	fun compare ((y1,x1,_), (y2,x2,_)) =
	    if      y1 > y2 then GREATER
	    else if y1 < y2 then LESS
	    else if x1 > x2 then GREATER
	    else if x1 < x2 then LESS
	    else raise KCsiga("Ismtld rcspont a(z) " ^ f ^ " llomnyban.")

	val istr    = TextIO.openIn f
	val meret   = valOf (Int.fromString (TextIO.inputLine istr))
	val ciklus  = valOf (Int.fromString (TextIO.inputLine istr))
	val adottak = Listsort.sort compare (adottak_be istr)
	val leiro   = (meret, ciklus, adottak) before TextIO.closeIn istr
    in
	if helyes_feladvany_leiro leiro then leiro
	else raise KCsiga("Hibs feladvnyler a(z) " ^ f ^ " llomnyban.")
    end

fun csigarajz m =
    let
	val s = length m

	fun next_row (y,_) = (y+1,1)
	fun next_col (y,x) = (y,x+1)

	fun field 0 = " - "
	  | field i = (if i < 10 then " " else "") ^ Int.toString i ^ " "

	(* vsep (y,x) = fgg. elvlaszt az (y,x) koordintj pont el *)
	fun vsep (y,x) =
	    if y <= (s+1) div 2 (* a csigavonal fels ill. als fele *)
	    then if x < y orelse x+y-1 > s  then "|" else " "
	    else if x > y orelse x+y-1 <= s then "|" else " "

	(* hsep (y,x) = vzsz. elvlaszt az (y,x) koordintj pont al *)
	fun hsep (y,x) =
	    if y <= s div 2 (* a csigavonal fels ill. als fele *)

	    then if x < y          then "|   "
	         else if x = y     then "/---"
		 else if x+y-1 < s then "----"
		 else if x+y-1 = s then "\\   "
		 else                   "|   "

	    else if x > y+1        then "|   "
	         else if x = y+1   then "/   "
		 else if x+y-1 > s then "----"
		 else if x+y-1 = s then "\\---"
		 else                   "|   "

	(* hsep_end y = sorzr az y. sor utni elvlaszt sorba *)
	fun hsep_end y =
	    if y = 0      then "\\\n"
	    else if y = s then "/\n"
	    else               "|\n"

	(* hsep_line y = vzsz. elvlaszt sor az y. sor utn *)
	fun hsep_line y =
	    let fun hsl (x, r) =
		if x > s then r ^ hsep_end y
		else hsl (x+1, r ^ hsep (y,x))
	    in
		hsl (1, "")
	    end
	
	fun csr ([], _, r) = r
	  | csr ([]::fss, c, r) =
	       csr(fss, next_row c, r ^ "|\n" ^ hsep_line (#1 c))
	  | csr ((f::fs)::fss, c, r) =
	       csr(fs::fss, next_col c, r ^ vsep c ^ field f)
    in
	csr(m, (1,1), hsep_line 0)
    end

(* csigarajzok ms = az ms megoldslistnak megfelel csigarajzok, res
   sorokkal elvlasztva
   csigarajzok : TCsiga.csigatabla.list -> string
 *)
fun csigarajzok ms = foldl (fn (m, s) => s ^ csigarajz m ^ "\n") "" ms

(* csiga_ki(m, ms) = az ms megoldslistt olvashat formban kirja az
   m szveges llomnyba.
   csiga_ki : string * TCsiga.csigatabla list -> unit
*)
fun csiga_ki(m, ms) =
      let val ostr = TextIO.openOut m
	  val _ = TextIO.output(ostr, csigarajzok ms)
      in
	 TextIO.closeOut ostr
      end
  
(* megold(f, m) = beolvas egy feladvnyt az f szveges llomnybl s kirja az
                  sszes megoldst az m szveges llomnyba. Ehhez termszetesen
                  szksge van a Csiga.buvos_csiga fggvnyre. Eredmnye az f
                  nevet, a megoldsok szmt s a futsi idt msodpercben
                  tartalmaz fzr.
   megold : string * string -> string
*)
fun megold (f, m) =
    let val fs = csiga_be f
	val start = Timer.startCPUTimer()
	val ms = Csiga.buvos_csiga fs
	val {usr=tim,...} = Timer.checkCPUTimer start
	val _ = csiga_ki(m, ms)
    in
	"Ler: " ^ f ^ ", " ^
	Int.toString(length ms) ^ " megolds, " ^
	"id: " ^ Time.fmt 2 tim ^ " mp\n"
    end

end
